from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from ...common import ConfigError
from .config_path import ConfPath

if TYPE_CHECKING:
    from .config_reference import ConfigReference
from .value_reference import ValueReference


@dataclass
class DefaultValueWrapper:
    """To distinguish between None and DefaultValueWrapper(None)."""

    value: Any


class DefaultValue(ABC):
    @abstractmethod
    def get_default(
        self,
        defaulter: ConfigDefaulter,
    ) -> DefaultValueWrapper | None: ...

    @abstractmethod
    def get_name(self) -> str: ...


class ConfigDefaulter:
    def __init__(
        self,
        root: ConfigReference,
        root_values: ValueReference,
        ref: ConfigReference | None = None,
        value_path: ConfPath | None = None,
    ) -> None:
        if ref is None:
            ref = root
        if value_path is None:
            value_path = ConfPath()
        self.root = root
        self.root_values = root_values
        self.ref = ref
        self.value_path = value_path

    def get_value_or_default(self) -> DefaultValueWrapper | None:
        try:
            return DefaultValueWrapper(self.root_values.get_value(self.value_path))
        except KeyError:
            return self.ref.config.default.get_default(self)

    def update_default(self):
        value_set = True
        if not self.root_values.is_value_set(self.value_path):
            default = self.ref.config.default.get_default(self)
            if default is not None:
                default_value = ValueReference(self.value_path, default.value)
                value = self.ref.config.verify(default_value)
                self.root_values.set_value(self.value_path, value)
            else:
                value_set = False
        if value_set:
            for name in self.ref.sub_options:
                ref = self.ref.sub_ref(name).resolve_inheritance(self.root)
                val_path = self.value_path.join(name)
                ConfigDefaulter(
                    root=self.root,
                    root_values=self.root_values,
                    ref=ref,
                    value_path=val_path,
                ).update_default()


class DefaultValueValue(DefaultValue):
    def __init__(self, value: Any) -> None:
        self.value = value

    def get_default(
        self,
        defaulter: ConfigDefaulter,
    ) -> DefaultValueWrapper | None:
        return DefaultValueWrapper(deepcopy(self.value))

    def get_name(self) -> str:
        if isinstance(self.value, bool):
            return str(self.value).lower()
        return repr(self.value)


class NoDefaultValue(DefaultValue):
    def __init__(self, name: str = "none"):
        self.name = name

    def get_default(
        self,
        defaulter: ConfigDefaulter,
    ) -> DefaultValueWrapper | None:
        return None

    def get_name(self) -> str:
        return self.name


class MissingDefaultError(ConfigError):
    """The user did not provide a required value."""


class RequiredValue(DefaultValue):
    def get_default(
        self,
        defaulter: ConfigDefaulter,
    ) -> DefaultValueWrapper | None:
        msg = f"{defaulter.value_path} requires a value"
        raise MissingDefaultError(msg)

    def get_name(self) -> str:
        return "required"


class RefDefaultValue(DefaultValue):
    def __init__(self, path: ConfPath, relative: bool = False) -> None:
        super().__init__()
        self.path: ConfPath = path
        self.relative = relative

    def get_default(
        self,
        defaulter: ConfigDefaulter,
    ) -> DefaultValueWrapper | None:
        if self.relative:
            rel_path = ConfPath(("^", *self.path.pth))
            target_path = defaulter.ref.config_path.join(rel_path)
            value_path = defaulter.value_path.join(rel_path)
        else:
            value_path = target_path = self.path
        try:
            target_opt = defaulter.root.sub_ref(target_path)
        except KeyError as e:
            msg = f"DefaultValue: reference to nonexisting option {target_path}"
            msg += f" (in {defaulter.ref.config_path})"
            raise ValueError(msg) from e
        return ConfigDefaulter(
            root=defaulter.root,
            root_values=defaulter.root_values,
            ref=target_opt.resolve_inheritance(defaulter.root),
            value_path=value_path,
        ).get_value_or_default()

    def get_name(self) -> str:
        r = str(self.path)
        if not self.relative:
            r = "/" + r
        return r



from __future__ import annotations

import logging
import tempfile
from pathlib import Path

from .build import _BuildBackend as std_backend
from .commands.cmd_runner import CommandRunner
from .common import ComponentConfig, PackageInfo, format_and_rethrow_exception
from .config import load as config_load
from .export import metadata as export_metadata

logger = logging.getLogger(__name__)


class _BuildComponentBackend:
    # --- Constructor ---------------------------------------------------------

    def __init__(self) -> None:
        self.runner: CommandRunner = CommandRunner()

    @property
    def verbose(self):
        return self.runner.verbose

    # --- Methods required by PEP 517 -----------------------------------------

    def get_requires_for_build_wheel(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-wheel"""
        try:
            self.parse_config_settings(config_settings)

            comp_source_dir = Path().resolve()
            comp_cfg = self.read_all_metadata(
                comp_source_dir, config_settings, self.verbose
            )
            cfg = std_backend.read_config(
                Path(comp_cfg.component["main_project"]),
                config_settings,
                self.verbose,
            )
            return std_backend.get_requires_build_project(
                config_settings, cfg, self.runner
            )
        except Exception as e:
            format_and_rethrow_exception(e)

    def get_requires_for_build_editable(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0660/#get-requires-for-build-editable"""
        return self.get_requires_for_build_wheel(config_settings)

    def get_requires_for_build_sdist(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-sdist"""
        return []

    def build_wheel(
        self, wheel_directory, config_settings=None, metadata_directory=None
    ):
        """https://www.python.org/dev/peps/pep-0517/#build-wheel"""
        try:
            assert metadata_directory is None

            # Parse options
            self.parse_config_settings(config_settings)

            # Build wheel
            with tempfile.TemporaryDirectory() as tmp_build_dir:
                return self.build_wheel_in_dir(
                    wheel_directory, tmp_build_dir, config_settings
                )
        except Exception as e:
            format_and_rethrow_exception(e)

    def build_editable(
        self, wheel_directory, config_settings=None, metadata_directory=None
    ):
        msg = "Editable installation not supported for individual components."
        raise NotImplementedError(msg)

    def build_sdist(self, sdist_directory, config_settings=None):
        msg = "Source distribution not supported for individual components."
        raise NotImplementedError(msg)

    # --- Parsing config options and metadata ---------------------------------

    def parse_config_settings(self, config_settings: dict | None):
        try:
            level = std_backend.get_log_level(config_settings)
            logging.basicConfig(level=level)
        except ValueError as e:
            logger.error("Invalid log level specified", exc_info=e)
        self.runner.verbose = std_backend.is_verbose_enabled(config_settings)

    @staticmethod
    def read_all_metadata(src_dir, config_settings, verbose):
        return config_load.read_full_component_config(
            src_dir / "pyproject.toml", config_settings, verbose
        )

    # --- Building wheels -----------------------------------------------------

    def build_wheel_in_dir(
        self, wheel_dir, tmp_build_dir, config_settings, editable=False
    ):
        """This is the main function that contains all steps necessary to build
        a complete wheel package, including the CMake builds etc."""
        comp_source_dir = Path().resolve()

        # Load metadata from the current (component) pyproject.toml file
        comp_cfg = self.read_all_metadata(
            comp_source_dir, config_settings, self.verbose
        )

        # Load the config from the main pyproject.toml file
        src_dir = Path(comp_cfg.component["main_project"]).resolve()
        cfg, module = std_backend.read_all_metadata(
            src_dir, config_settings, self.verbose
        )
        pkg_info = std_backend.get_pkg_info(comp_cfg, module)
        cmake_cfg = std_backend.get_cmake_config(cfg)

        # Set up all paths
        paths = std_backend.get_default_paths(
            wheel_dir, tmp_build_dir, src_dir, cfg, cmake_cfg
        )

        # Create dist-info folder
        distinfo_dir = f"{pkg_info.norm_name}-{pkg_info.version}.dist-info"
        distinfo_dir = paths.pkg_staging_dir / distinfo_dir
        distinfo_dir.mkdir(parents=True, exist_ok=True)

        # Write metadata, license and entry points to Wheel's distinfo
        export_metadata.write_metadata(comp_cfg, distinfo_dir)
        export_metadata.write_license_files(comp_cfg, distinfo_dir)
        export_metadata.write_entry_points(comp_cfg, distinfo_dir)

        # Configure, build and install the CMake project
        if cmake_cfg:
            cmaker = self.get_cmaker(
                paths.source_dir,
                paths.build_dir,
                paths.staging_dir,
                cmake_cfg,
                cfg.cross,
                pkg_info,
                comp_cfg,
                runner=self.runner,
            )
            if not comp_cfg.component["install_only"]:
                cmaker.build()
            cmaker.install()

        # Create wheel
        return std_backend.create_wheel(paths, cfg, cmake_cfg, pkg_info)

    # --- CMake builds --------------------------------------------------------

    @staticmethod
    def get_cmaker(
        source_dir: Path,
        build_dir: Path,
        install_dir: Path | None,
        cmake_cfg: dict,
        cross_cfg: dict | None,
        package_info: PackageInfo,
        comp_cfg: ComponentConfig,
        **kwargs,
    ):
        cmaker = std_backend.get_cmaker(
            source_dir=source_dir,
            build_dir=build_dir,
            install_dir=install_dir,
            cmake_cfg=cmake_cfg,
            cross_cfg=cross_cfg,
            package_info=package_info,
            **kwargs,
        )
        comp = comp_cfg.component
        if "build_presets" in comp:
            cmaker.build_settings.presets = comp["build_presets"]
        if "build_args" in comp:
            cmaker.build_settings.args = comp["build_args"]
        if "build_tool_args" in comp:
            cmaker.build_settings.tool_args = comp["build_tool_args"]
        if "install_args" in comp:
            cmaker.install_settings.args = comp["install_args"]
        if "install_components" in comp:
            cmaker.install_settings.components = comp["install_components"]
        return cmaker


_BACKEND = _BuildComponentBackend()
get_requires_for_build_wheel = _BACKEND.get_requires_for_build_wheel
get_requires_for_build_sdist = _BACKEND.get_requires_for_build_sdist
get_requires_for_build_editable = _BACKEND.get_requires_for_build_editable
build_wheel = _BACKEND.build_wheel
build_sdist = _BACKEND.build_sdist
build_editable = _BACKEND.build_editable



from __future__ import annotations

import logging
import os
import sys
import sysconfig
from dataclasses import dataclass
from pathlib import Path
from string import Template

from .. import __version__
from ..common import PackageInfo
from ..common.util import python_sysconfig_platform_to_cmake_platform_win
from .cmd_runner import CommandRunner

logger = logging.getLogger(__name__)


@dataclass
class CMakeSettings:
    working_dir: Path
    source_path: Path
    build_path: Path
    os: str
    find_python: bool
    find_python3: bool
    minimum_required: str
    command: Path = Path("cmake")


@dataclass
class CMakeConfigureSettings:
    environment: dict
    build_type: str | None
    options: dict[str, str]
    args: list[str]
    preset: str | None
    generator: str | None
    cross_compiling: bool
    toolchain_file: Path | None
    python_prefix: Path | None
    python_library: Path | None
    python_include_dir: Path | None
    python_interpreter_id: str | None


@dataclass
class CMakeBuildSettings:
    args: list[str]
    tool_args: list[str]
    presets: list[str]
    configs: list[str]


@dataclass
class CMakeInstallSettings:
    args: list[str]
    configs: list[str]
    components: list[str]
    prefix: Path | None


@dataclass
class Option:
    name: str
    value: str
    type: str = "STRING"
    description: str = ""


class CMaker:
    def __init__(
        self,
        cmake_settings: CMakeSettings,
        conf_settings: CMakeConfigureSettings,
        build_settings: CMakeBuildSettings,
        install_settings: CMakeInstallSettings,
        package_info: PackageInfo,
        runner: CommandRunner,
    ):
        self.cmake_settings = cmake_settings
        self.conf_settings = conf_settings
        self.build_settings = build_settings
        self.install_settings = install_settings
        self.package_info = package_info
        self.runner = runner
        self.environment: dict[str, str] | None = None

    def run(self, *args, **kwargs):
        return self.runner.run(*args, **kwargs)

    def prepare_environment(self):
        """Copy of the current environment with the variables defined in the
        user's configuration settings."""
        if self.environment is None:
            self.environment = os.environ.copy()
            self.environment["PY_BUILD_CMAKE_VERSION"] = str(__version__)
            self.environment["PY_BUILD_CMAKE_BINARY_DIR"] = str(
                self.cmake_settings.build_path
            )
            if self.install_settings.prefix is not None:
                self.environment["PY_BUILD_CMAKE_INSTALL_PREFIX"] = str(
                    self.install_settings.prefix
                )
            if self.conf_settings.environment:
                for k, v in self.conf_settings.environment.items():
                    templ = Template(v)
                    self.environment[k] = templ.substitute(self.environment)
        return self.environment

    def cross_compiling(self) -> bool:
        return self.conf_settings.cross_compiling

    def get_configure_options_package(self) -> list[Option]:
        """Flags specific to py-build-cmake, useful in the user's CMake scripts."""
        return [
            Option("PY_BUILD_CMAKE_VERSION", str(__version__)),
            Option("PY_BUILD_CMAKE_PACKAGE_VERSION", self.package_info.version),
            Option("PY_BUILD_CMAKE_PACKAGE_NAME", self.package_info.norm_name),
            Option("PY_BUILD_CMAKE_MODULE_NAME", self.package_info.module_name),
        ]

    def get_native_python_prefixes(self) -> str:
        """Get the prefix paths to locate this (native) Python installation in,
        as a semicolon-separated string."""
        pfxs = [
            Path(sys.exec_prefix).as_posix(),
            Path(sys.prefix).as_posix(),
            Path(sys.base_exec_prefix).as_posix(),
            Path(sys.base_prefix).as_posix(),
        ]
        return ";".join(pfxs)

    def get_native_python_abi_tuple(self):
        abiflag = lambda c: c in sys.abiflags
        onoff = lambda x: "ON" if x else "OFF"
        return ";".join(map(onoff, map(abiflag, "dmu")))

    def get_native_python_implementation(self) -> str | None:
        return {
            "cpython": "CPython",
            "pypy": "PyPy",
        }.get(sys.implementation.name)

    def get_native_python_hints(self, prefix):
        """FindPython hints and artifacts for this (native) Python installation."""
        yield Option(prefix + "_ROOT_DIR", self.get_native_python_prefixes())
        impl = self.get_native_python_implementation()
        if impl:
            yield Option(prefix + "_FIND_IMPLEMENTATIONS", impl)
        return
        # FIND_ABI seems to confuse CMake
        yield Option(prefix + "_FIND_ABI", self.get_native_python_abi_tuple())
        if impl == "PyPy":
            yield Option(prefix + "_INCLUDE_DIR", sysconfig.get_path("platinclude"))

    def get_cross_python_hints(self, prefix):
        """FindPython hints and artifacts to set when cross-compiling."""
        if self.conf_settings.python_prefix:
            pfx = self.conf_settings.python_prefix.as_posix()
            yield Option(prefix + "_ROOT_DIR", pfx, "PATH")
        if self.conf_settings.python_library:
            lib = self.conf_settings.python_library.as_posix()
            yield Option(prefix + "_LIBRARY", lib, "FILEPATH")
        if self.conf_settings.python_include_dir:
            inc = self.conf_settings.python_include_dir.as_posix()
            yield Option(prefix + "_INCLUDE_DIR", inc, "PATH")
        if self.conf_settings.python_interpreter_id:
            id = self.conf_settings.python_interpreter_id
            yield Option(prefix + "_INTERPRETER_ID", id, "STRING")

    def get_configure_options_python(self) -> list[Option]:
        """Flags to help CMake find the right version of Python."""

        def get_opts(prefix):
            executable = Path(sys.executable).as_posix()
            yield Option(prefix + "_EXECUTABLE", executable, "FILEPATH")
            yield Option(prefix + "_FIND_REGISTRY", "NEVER")
            yield Option(prefix + "_FIND_FRAMEWORK", "NEVER")
            yield Option(prefix + "_FIND_STRATEGY", "LOCATION")
            yield Option(prefix + "_FIND_VIRTUALENV", "FIRST")
            if not self.cross_compiling():
                yield from self.get_native_python_hints(prefix)
            else:
                yield from self.get_cross_python_hints(prefix)

        opts = []
        if self.cmake_settings.find_python:
            opts += list(get_opts("Python"))
        if self.cmake_settings.find_python3:
            opts += list(get_opts("Python3"))
        return opts

    def get_configure_options_install(self) -> list[Option]:
        prefix = self.install_settings.prefix
        if prefix:
            return [
                Option("CMAKE_INSTALL_PREFIX", prefix.as_posix(), "PATH"),
                Option("CMAKE_FIND_NO_INSTALL_PREFIX", "On", "BOOL"),
            ]
        return []

    def get_configure_options_toolchain(self) -> list[str]:
        """Sets CMAKE_TOOLCHAIN_FILE."""
        return (
            ["CMAKE_TOOLCHAIN_FILE:FILEPATH=" + str(self.conf_settings.toolchain_file)]
            if self.conf_settings.toolchain_file
            else []
        )

    def get_configure_options_settings(self) -> list[str]:
        return [k + "=" + v for k, v in self.conf_settings.options.items()]

    def get_configure_options(self) -> list[str]:
        """Get the list of options (-D) passed to the CMake configure step
        through the command line."""
        return (
            self.get_configure_options_toolchain()
            + self.get_configure_options_settings()
        )

    def get_preload_options(self) -> list[Option]:
        """Get the list of options set in the CMake pre-load script (-C)."""
        return (
            self.get_configure_options_package()
            + self.get_configure_options_python()
            + self.get_configure_options_install()
        )

    def write_preload_options(self) -> list[str]:
        """Write the options into the CMake pre-load script and return the
        command-line flags that tell CMake to load it."""
        opts = self.get_preload_options()
        if not opts:
            return []

        def fmt_opt(o: Option):
            return (
                f'set({o.name} "{o.value}"'
                f' CACHE {o.type} "{o.description}" FORCE)\n'
            )

        preload_file = self.cmake_settings.build_path / "py-build-cmake-preload.cmake"
        version = self.cmake_settings.minimum_required
        if self.runner.verbose:
            print("Writing CMake pre-load file")
            print(f"{preload_file}")
            print("---------------------------")
            print(f"cmake_minimum_required(VERSION {version})")
            for o in opts:
                print(fmt_opt(o), end="")
            print("---------------------------\n")
        if not self.runner.dry:
            self.cmake_settings.build_path.mkdir(parents=True, exist_ok=True)
            with preload_file.open("w", encoding="utf-8") as f:
                f.write(f"cmake_minimum_required(VERSION {version})\n")
                for o in opts:
                    f.write(fmt_opt(o))
        return ["-C", str(preload_file)]

    def get_cmake_generator_platform(self) -> list[str]:
        win = self.cmake_settings.os == "windows"
        gen = self.conf_settings.generator
        vs_gen = win and (not gen or gen.lower().startswith("visual studio"))
        if vs_gen and not self.cross_compiling():
            plat = sysconfig.get_platform()
            cmake_plat = python_sysconfig_platform_to_cmake_platform_win(plat)
            if cmake_plat:
                return ["-A", cmake_plat]
            else:
                msg = "Unknown platform, CMake generator platform option (-A) will not be set"
                logger.warning(msg)
        return []

    def get_configure_command(self):
        options = self.get_configure_options()
        cmd = [str(self.cmake_settings.command)]
        cmd += ["-S", str(self.cmake_settings.source_path)]
        if self.conf_settings.preset:
            cmd += ["--preset", self.conf_settings.preset]
        if not self.build_settings.presets:
            cmd += ["-B", str(self.cmake_settings.build_path)]
        if self.conf_settings.generator:
            cmd += ["-G", self.conf_settings.generator]
        cmd += self.get_cmake_generator_platform()
        cmd += self.write_preload_options()
        cmd += [f for opt in options for f in ("-D", opt)]
        cmd += self.conf_settings.args
        return cmd

    def get_working_dir(self):
        cwd = self.cmake_settings.working_dir
        return str(cwd) if cwd is not None else None

    def configure(self):
        env = self.prepare_environment()
        cmd = self.get_configure_command()
        cwd = self.get_working_dir()
        self.run(cmd, cwd=cwd, check=True, env=env)

    def get_build_command(self, config, preset):
        cmd = [str(self.cmake_settings.command), "--build"]
        if preset is not None:
            cmd += ["--preset", preset]
        else:
            cmd += [str(self.cmake_settings.build_path)]
        if config is not None:
            cmd += ["--config", config]
        if self.build_settings.args:
            cmd += self.build_settings.args
        if self.build_settings.tool_args:
            cmd += ["--", *self.build_settings.tool_args]
        return cmd

    def get_build_commands(self):
        if self.build_settings.presets and self.build_settings.configs:
            msg = "Mixing both CMake build presets and plain configs is not recommended"
            logger.warning(msg)
        for preset in self.build_settings.presets:
            yield self.get_build_command(None, preset)
        for config in self.build_settings.configs:
            yield self.get_build_command(config, None)
        if not self.build_settings.presets and not self.build_settings.configs:
            yield self.get_build_command(None, None)

    def build(self):
        env = self.prepare_environment()
        cwd = self.get_working_dir()
        for cmd in self.get_build_commands():
            self.run(cmd, cwd=cwd, check=True, env=env)

    def get_install_command(self, config):
        for component in self.install_settings.components:
            cmd = [str(self.cmake_settings.command), "--install"]
            cmd += [str(self.cmake_settings.build_path)]
            if self.install_settings.prefix:
                cmd += ["--prefix", str(self.install_settings.prefix)]
            if config is not None:
                cmd += ["--config", config]
            if component:
                cmd += ["--component", component]
            if self.install_settings.args:
                cmd += self.install_settings.args
            yield cmd

    def get_install_commands(self):
        for config in self.install_settings.configs:
            yield from self.get_install_command(config)
        if not self.install_settings.configs:
            yield from self.get_install_command(None)

    def install(self):
        env = self.prepare_environment()
        cwd = self.get_working_dir()
        for cmd in self.get_install_commands():
            self.run(cmd, cwd=cwd, check=True, env=env)



from __future__ import annotations

from copy import copy

from ...common import ConfigError
from .config_option import ConfigOption
from .config_path import ConfPath
from .default import DefaultValue
from .value_reference import ValueReference


class ListOfStrConfigOption(ConfigOption):
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
        convert_str_to_singleton=False,
        append_by_default=False,
    ) -> None:
        super().__init__(
            name,
            description,
            example,
            default,
            inherit_from,
            create_if_inheritance_target_exists,
        )
        self.convert_str_to_singleton = convert_str_to_singleton
        self.append_by_default = append_by_default

    list_op_keys = frozenset(("+", "-", "=", "value", "append", "prepend"))

    def get_typename(self, md: bool = False):
        return "list+" if self.append_by_default else "list"

    def _override_list(self, old_value: ValueReference, new_value: ValueReference):
        assert isinstance(new_value.values, list)
        if self.append_by_default:
            assert isinstance(old_value.values, list)
            return copy(old_value.values + new_value.values)
        else:
            return copy(new_value.values)

    def _override_dict(self, old_value: ValueReference, new_value: ValueReference):
        assert isinstance(new_value.values, dict)
        if "=" in new_value.values:
            return copy(new_value.values["="])
        if "value" in new_value.values:
            return copy(new_value.values["value"])
        result = copy(old_value.values)
        if "-" in new_value.values:
            remove = set(new_value.values["-"])
            result = [v for v in result if v not in remove]
        if "+" in new_value.values:
            result = result + new_value.values["+"]
        if "append" in new_value.values:
            result = result + new_value.values["append"]
        if "prepend" in new_value.values:
            result = new_value.values["prepend"] + result
        return result

    def override(self, old_value: ValueReference, new_value: ValueReference):
        if old_value.values is None:
            old_value.values = []
        if new_value.values is None:
            return old_value.values
        if not isinstance(old_value.values, list):
            msg = f"Type of {old_value.value_path} should be {list}, "
            msg += f"not {type(old_value.values)}"
            raise ConfigError(msg)
        if isinstance(new_value.values, dict):
            return self._override_dict(old_value, new_value)
        return self._override_list(old_value, new_value)

    def verify(self, values: ValueReference):
        if isinstance(values.values, dict):
            invalid_keys = set(values.values.keys()) - self.list_op_keys
            if invalid_keys:
                inv_str = ", ".join(map(str, invalid_keys))
                val_str = ", ".join(map(str, self.list_op_keys))
                msg = f"Invalid keys in {values.value_path}: {inv_str} "
                msg += f"(valid keys are: {val_str})"
                raise ConfigError(msg)
            for k, v in values.values.items():
                pthname = f"{values.value_path}[{k}]"
                if not isinstance(v, list):
                    msg = f"Type of {pthname} should be {list}, not {type(v)}"
                    raise ConfigError(msg)
                if not all(isinstance(el, str) for el in v):
                    msg = f"Type of elements in {pthname} should be {str}"
                    raise ConfigError(msg)
        elif not isinstance(values.values, list):
            if self.convert_str_to_singleton and isinstance(values.values, str):
                values.values = [values.values]
            else:
                msg = f"Type of {values.value_path} should be {list}, "
                msg += f"not {type(values.values)}"
                raise ConfigError(msg)
        elif not all(isinstance(el, str) for el in values.values):
            msg = f"Type of elements in {values.value_path} should be {str}"
            raise ConfigError(msg)
        return values.values

    def finalize(self, values: ValueReference):
        if isinstance(values.values, str):
            return [values.values]
        return values.values



from __future__ import annotations

import shutil
from pathlib import Path

from ..common import Module


def copy_pkg_source_to(staging_dir: Path, module: Module, symlink: bool = False):
    """Copy the files of a Python package to the build directory."""
    for src in module.iter_files_abs():
        rel_path = src.relative_to(module.prefix)
        dst = staging_dir / rel_path
        dst.parent.mkdir(parents=True, exist_ok=True)
        if symlink:
            dst.symlink_to(src, target_is_directory=False)
        else:
            shutil.copy2(src, dst, follow_symlinks=False)



"""
The following functions and classes are based on flit_core, under the BSD 3-Clause license:

Copyright (c) 2015, Thomas Kluyver and contributors
All rights reserved.

BSD 3-clause license:

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import annotations

import io
import logging
import os
import tarfile
from copy import copy
from gzip import GzipFile
from pathlib import Path, PurePosixPath
from typing import Iterable

from pyproject_metadata import StandardMetadata

from ..common import ConfigError, Module, PackageInfo

logger = logging.getLogger(__name__)


class SdistError(ConfigError):
    """Problem packaging the project's sdist"""


def normalize_file_permissions(st_mode):
    """Normalize the permission bits in the st_mode field from stat to 644/755

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755
    return new_mode


def clean_tarinfo(ti: tarfile.TarInfo, mtime=None) -> tarfile.TarInfo:
    """Clean metadata from a TarInfo object to make it more reproducible.

    - Set uid & gid to 0
    - Set uname and gname to ""
    - Normalise permissions to 644 or 755
    - Set mtime if not None
    """
    ti = copy(ti)
    ti.uid = 0
    ti.gid = 0
    ti.uname = ""
    ti.gname = ""
    ti.mode = normalize_file_permissions(ti.mode)
    if mtime is not None:
        ti.mtime = mtime
    return ti


class FilePatterns:
    """Manage a set of file inclusion/exclusion patterns relative to basedir"""

    def __init__(self, patterns: Iterable[str], basedir: Path):
        self.basedir = basedir

        self.dirs = set()
        self.files = set()

        for pattern in patterns:
            for path in sorted(basedir.glob(pattern)):
                rel = path.relative_to(basedir)
                if path.is_dir():
                    self.dirs.add(rel)
                else:
                    self.files.add(rel)

    def match_file(self, rel_path: Path) -> bool:
        if rel_path in self.files:
            return True

        # Check if it's contained in any directory in the list
        return any(d in rel_path.parents for d in self.dirs)

    def match_dir(self, rel_path: Path) -> bool:
        if rel_path in self.dirs:
            return True

        # Check if it's a subdirectory of any directory in the list
        return any(d in rel_path.parents for d in self.dirs)


class SdistBuilder:
    """Builds a minimal sdist

    These minimal sdists should work for PEP 517.
    The class is extended in flit.sdist to make a more 'full fat' sdist,
    which is what should normally be published to PyPI.
    """

    def __init__(
        self,
        module: Module,
        pkg_info: PackageInfo,
        metadata: StandardMetadata,
        cfgdir: Path,
        extra_files,
        include_patterns: Iterable[str] = (),
        exclude_patterns: Iterable[str] = (),
    ):
        self.module = module
        self.pkg_info = pkg_info
        self.metadata = metadata
        self.cfgdir = cfgdir
        self.extra_files = extra_files
        self.includes = FilePatterns(include_patterns, cfgdir)
        self.excludes = FilePatterns(exclude_patterns, cfgdir)

    def select_files(self):
        """Pick which files from the source tree will be included in the sdist

        This is overridden in flit itself to use information from a VCS to
        include tests, docs, etc. for a 'gold standard' sdist.
        """
        make_rel = lambda p: p.relative_to(self.module.base_path)
        yield from map(make_rel, self.module.iter_files_abs())
        yield from map(make_rel, self.extra_files)

    def crucial_files(self):
        make_rel = lambda p: p.relative_to(self.module.base_path)
        if not self.module.is_generated and not self.module.is_namespace:
            yield make_rel(self.module.full_file)
        yield from map(make_rel, self.extra_files)

    def apply_includes_excludes(self, files: Iterable[Path]):
        files = {f for f in files if not self.excludes.match_file(f)}

        for f_rel in self.includes.files:
            if not self.excludes.match_file(f_rel):
                files.add(Path(f_rel))

        for rel_d in self.includes.dirs:
            for dirpath, dirs, dfiles in os.walk(self.cfgdir / rel_d):
                file: str
                for file in dfiles:
                    f_abs = Path(dirpath) / file
                    f_rel = f_abs.relative_to(self.cfgdir)
                    if not self.excludes.match_file(f_rel):
                        files.add(f_rel)

                # Filter subdirectories before os.walk scans them
                dirs[:] = [
                    d
                    for d in dirs
                    if not self.excludes.match_dir(
                        (Path(dirpath) / d).relative_to(self.cfgdir)
                    )
                ]

        crucial_files = set(self.crucial_files())
        missing_crucial = crucial_files - files
        if missing_crucial:
            msg = f"Crucial files were excluded from the sdist: {', '.join(map(str, missing_crucial))}"
            raise SdistError(msg)

        return sorted(files)

    def add_setup_py(self, files_to_add, target_tarfile):
        """No-op here; overridden in flit to generate setup.py"""

    @property
    def dir_name(self):
        return f"{self.pkg_info.norm_name}-{self.pkg_info.version}"

    def build(self, target_dir: Path, gen_setup_py=True):
        target_dir.mkdir(parents=True, exist_ok=True)
        target = target_dir / (self.dir_name + ".tar.gz")
        source_date_epoch = os.environ.get("SOURCE_DATE_EPOCH", "")
        mtime = int(source_date_epoch) if source_date_epoch else None
        gz = GzipFile(str(target), mode="wb", mtime=mtime)
        tf = tarfile.TarFile(
            str(target), mode="w", fileobj=gz, format=tarfile.PAX_FORMAT
        )

        try:
            files_to_add = self.apply_includes_excludes(self.select_files())
            archive_dir = PurePosixPath(self.dir_name)
            for relpath in files_to_add:
                path = self.cfgdir / relpath
                archive_path = archive_dir / PurePosixPath(relpath)
                ti = tf.gettarinfo(path, arcname=str(archive_path))
                ti = clean_tarinfo(ti, mtime)

                if ti.isreg():
                    with path.open("rb") as f:
                        tf.addfile(ti, f)
                else:
                    tf.addfile(ti)  # Symlinks & ?

            if gen_setup_py:
                self.add_setup_py(files_to_add, tf)

            stream = io.StringIO()
            stream.write(str(self.metadata.as_rfc822()))
            pkg_info = stream.getvalue().encode()
            ti = tarfile.TarInfo(str(archive_dir / "PKG-INFO"))
            ti.size = len(pkg_info)
            tf.addfile(ti, io.BytesIO(pkg_info))

        finally:
            tf.close()
            gz.close()

        logger.debug("Built sdist: %s", target)
        return target
